/*-
 * #%L
 * ComDet Plugin for ImageJ
 * %%
 * Copyright (C) 2012 - 2023 Cell Biology, Neurobiology and Biophysics
 * Department of Utrecht University.
 * %%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-3.0.html>.
 * #L%
 */
package fiji.plugin.ComDet;

import java.awt.AWTEvent;
import java.awt.Color;

import ij.CompositeImage;
import ij.IJ;
import ij.ImagePlus;
import ij.Prefs;
import ij.gui.DialogListener;
import ij.gui.GenericDialog;
import ij.gui.Overlay;
import ij.gui.Roi;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;


public class CDDialog implements DialogListener{

	
	/** current active ImagePlus **/
	public ImagePlus imp;
	
	
	CDAnalysis cd;
	/**image width of original image**/
	int nWidth;  
	/**image height of original image**/
	int nHeight; 
	/**number of slices**/
	int nSlices; 
	
	/** original overlay **/
	Overlay savedOverlay;
	
	
	//finding particles
	/**standard deviation of PSF approximated by Gaussian**/
	double [] dPSFsigma; 
	/**size of Gaussian kernel for particles enhancement**/
	int [] nKernelSize; 
	/**number of threads for calculation**/
	int nThreads; 
	/**threshold of minimum particle area for each channel **/
	int [] nAreaCut; 
	/**threshold of maximum particle area for each channel **/
	int [] nAreaMax; 
	/**sensitivity of detection for each channel **/
	float [] nSensitivity;
	/** whether to include big particles in the detection (for each channel) **/
	boolean [] bBigParticles;
	/** Whether to segment large particles (for each channel) **/
	boolean [] bSegmentLargeParticles;	
	/** ROI of what shape to add 0 = ovals, 1 = rectangles **/
	int nRoiOption;
	
	/** number of channels **/
	public int ChNumber;
	/** current channel in processing (starting from 1!) **/
	int iCh;
	
	/** main dialog window **/
	GenericDialog fpDial;
	
	boolean bMultiChannelDetection;

	
	
	/** dialog options**/
	String [] sROIManagerOneCh;
	String [] sROIManagerMultiCh;
	
	String [] sSummaryOptions;
	/** 0 reset summary table; 1 append to a summary table **/
	int nSummaryOptions;
	
	/** Flag determing whether to add detections to ROI Manager
	 * 0 - do not add
	 * 1 - add all detections
	 * 2 - add only colocalized particles
	 * 3 - add only non-colocalized particles
	 * **/
	int nRoiManagerAdd;
	/** ROIs of what shape to add **/
	String [] sRoiOptions; 

	
	//boolean bPreview;

	
	//colocalization analysis parameters
	boolean bColocalization;
	double dColocDistance;
	boolean bPlotMultiChannels;
	
	/** default constructor **/ 
	public CDDialog()
	{
		//dPSFsigma = new double[2];
		//nKernelSize = new int[2];
		//nAreaCut = new int[2];
		//nAreaMax =  new int[2];
		//nSensitivity = new float[2];
		sROIManagerOneCh = new String [] {
				"Nothing", "All detections"};
		sROIManagerMultiCh = new String [] {
				"Nothing", "All detections","Only colocalized particles", "Only non-colocalized particles"};
		sRoiOptions = new String [] {"Ovals","Rectangles"};
		
		sSummaryOptions =new String [] {"Reset","Append"};
		
	}
	/** function initializing dialog for particles detection,
	 * reads parameters of current image (or Stack/HyperStack, etc) **/
	public void initImage(ImagePlus imp_in, int [] imageinfo_in)
	{
		imp=imp_in;
		ChNumber=imageinfo_in[2];
		dPSFsigma = new double[ChNumber];
		nKernelSize = new int[ChNumber];
		nAreaCut = new int[ChNumber];
		nAreaMax =  new int[ChNumber];
		nSensitivity = new float[ChNumber];
		bBigParticles = new boolean[ChNumber];
		bSegmentLargeParticles = new boolean[ChNumber];	
		
	}

	//dialog showing options for particle search algorithm		
	public boolean findParticles() {
		
		String nCurrentVersion;

		
		//check if it is new version
		// if yes, display one time message
		nCurrentVersion=Prefs.get("ComDet.PluginVersion","none");
		if(!nCurrentVersion.equals(ComDetConstants.ComDetVersion))
		{
			IJ.showMessage("This is a new installation or an update to "+ComDetConstants.ComDetVersion+" version of ComDet plugin!\n "
					//+ "Take notice that detection algorithm has changed in comparison to previous versions.\n"
					+"Check https://github.com/ekatrukha/ComDet/wiki/Updates-history for description of changes. ");
			Prefs.set("ComDet.PluginVersion",ComDetConstants.ComDetVersion);
		}
		
		//multichannel input, ask user for the parameters
		if(ChNumber>1)
		{
			fpDial = new GenericDialog("Detect Particles");
			fpDial.addMessage("Multi-channel image is detected as input. \n");
			fpDial.addCheckbox("Calculate colocalization?", Prefs.get("ComDet.bColocalization", false));
			fpDial.addMessage("Colocalization analysis parameters:\n");
			fpDial.addNumericField("Max distance between colocalized spots", Prefs.get("ComDet.dColocDistance", 4), 2,5," pixels");
			fpDial.addCheckbox("Plot detected particles in all channels?", Prefs.get("ComDet.bPlotMultiChannels", false));
			fpDial.addChoice("ROIs shape: ",sRoiOptions, Prefs.get("ComDet.nRoiOption", "Ovals"));
			fpDial.addChoice("Add to ROI Manager:", sROIManagerMultiCh, Prefs.get("ComDet.sROIManagerMulti", "Nothing"));
			fpDial.addChoice("Summary Table:", sSummaryOptions, Prefs.get("ComDet.sSummaryOptions", "Reset"));
			
			
			fpDial.showDialog();
			if (fpDial.wasCanceled())
	            return false;
			bColocalization = fpDial.getNextBoolean();
			Prefs.set("ComDet.bColocalization", bColocalization);			
			dColocDistance = fpDial.getNextNumber();
			Prefs.set("ComDet.dColocDistance", dColocDistance);	
			bPlotMultiChannels = fpDial.getNextBoolean();
			Prefs.set("ComDet.bPlotMultiChannels", bPlotMultiChannels);
			nRoiOption = fpDial.getNextChoiceIndex();
			Prefs.set("ComDet.nRoiOption", sRoiOptions[nRoiOption]);			
			nRoiManagerAdd = fpDial.getNextChoiceIndex();
			Prefs.set("ComDet.sROIManagerMulti", sROIManagerMultiCh[nRoiManagerAdd]);
			nSummaryOptions = fpDial.getNextChoiceIndex();
			Prefs.set("ComDet.sSummaryOptions", sSummaryOptions[nSummaryOptions]);
			
			if(!bColocalization && nRoiManagerAdd>=2)
			{
				nRoiManagerAdd=0;
				IJ.log("Cannot add colocalized particles to ROI, since colocalization option is unchecked. Nothing will be added.");
			}
			Prefs.set("ComDet.sROIManagerMultiCh", sROIManagerMultiCh[nRoiManagerAdd]);
		}

		savedOverlay=imp.getOverlay();
		
		
		for (iCh=1;iCh<=ChNumber;iCh++)
		{
			String sChN = Integer.toString(iCh);
			fpDial = new GenericDialog("Detect Particles channel"+sChN);
			fpDial.addMessage("Detection parameters:\n");
			fpDial.addCheckbox("ch"+sChN+"i: Include larger particles?", Prefs.get("ComDet.bBigParticles"+sChN, true));	
			fpDial.addCheckbox("ch"+sChN+"l: Segment larger particles (slow)?", Prefs.get("ComDet.bSegmentLargeParticles"+sChN, false));
			fpDial.addNumericField("ch"+sChN+"a: Approximate particle size", Prefs.get("ComDet.dPSFsigma"+sChN, 4), 2,5," pixels");
			fpDial.addNumericField("ch"+sChN+"s: Intensity threshold (in SD):", Prefs.get("ComDet.dSNRT"+sChN, 3), 2,5, "around (3-20)");
			if(ChNumber == 1)
			{
				fpDial.addChoice("ROIs shape: ",sRoiOptions, Prefs.get("ComDet.nRoiOption", "Ovals"));
				fpDial.addChoice("Add to ROI Manager:", sROIManagerOneCh, Prefs.get("ComDet.sROIManagerOne", "Nothing"));
				fpDial.addChoice("Summary Table:", sSummaryOptions, Prefs.get("ComDet.sSummaryOptions", "Reset"));
			}

			ImagePlus fakeimp=new ImagePlus("quick",new FloatProcessor(1,1));
			fakeimp.show();
			fpDial.addPreviewCheckbox(null,"Preview detection..");
			fakeimp.changes=false;
			fakeimp.close();
			fpDial.addDialogListener(this);

			fpDial.showDialog();

			imp.setOverlay(savedOverlay);
			imp.updateAndRepaintWindow();
			imp.show();
			
			if (fpDial.wasCanceled())
				{return false;}

			setPrefs();

		}
		
		
		
		return true;
	}

	@Override
	public boolean dialogItemChanged(GenericDialog gd, AWTEvent e) {
		// TODO Auto-generated method stub
		
		CompositeImage multichannel_imp;
		int[] nImagePos;
		Roi RoiSelected;// = imp.getRoi();
		ImageProcessor ip;
		boolean bValuesOk;
		bValuesOk=getValues();
		if(!bValuesOk)
			return false;
		if(gd.wasOKed())
		{
			return true;
		}
		if(gd.isPreviewActive())
		{
				gd.previewRunning(true);
				cd = new CDAnalysis(ChNumber);
				
				if(ChNumber>1)
				{
					multichannel_imp=(CompositeImage) imp;
					multichannel_imp.setC(iCh);
					cd.colorCh= multichannel_imp.getChannelColor();
				}
				else
				{
					cd.colorCh=Color.YELLOW;
				}
				
				cd.overlay_= new Overlay();
				cd.initConvKernel(this,iCh-1);
				RoiSelected= imp.getRoi();
				//nImagePos = imp.convertIndexToPosition(imp.getSlice());
				nImagePos = new int[3];
				nImagePos[0]=iCh;
				nImagePos[1]=imp.getSlice();
				nImagePos[2]=imp.getFrame();
				imp.setPositionWithoutUpdate(nImagePos[0], nImagePos[1],nImagePos[2]);
				ip = imp.getProcessor().duplicate();
				
				cd.detectParticles(ip, this,nImagePos, 1, RoiSelected);
				imp.setOverlay(cd.overlay_);
				imp.updateAndRepaintWindow();
				imp.show();
				gd.previewRunning(false);
				
		}
		else
		{
			imp.setOverlay(savedOverlay);
			imp.updateAndRepaintWindow();
			imp.show();
		}
		return true;
	}
	

	/** function reads parameters values from the dialog **/
	public boolean getValues()
	{
		int indCh=iCh-1;

		bBigParticles[indCh] = fpDial.getNextBoolean();
		bSegmentLargeParticles[indCh] = fpDial.getNextBoolean();

		dPSFsigma[indCh] = fpDial.getNextNumber();
		if(Double.isNaN(dPSFsigma[indCh]))
			return false;
		
		nSensitivity[indCh] = (float)fpDial.getNextNumber();
		if(Double.isNaN(nSensitivity[indCh]))
			return false;
		
		
		
		if(ChNumber == 1)
		{
			nRoiOption = fpDial.getNextChoiceIndex();
			nRoiManagerAdd = fpDial.getNextChoiceIndex();
			nSummaryOptions = fpDial.getNextChoiceIndex();
			
		}
		//bPreview = fpDial.getNextBoolean();
		nThreads = 50;

			
		dPSFsigma[indCh] *= 0.5;
		nAreaMax[indCh] = (int) (12.0*dPSFsigma[indCh]*dPSFsigma[indCh]);
		//putting limiting criteria on spot size
		nAreaCut[indCh] = (int) (dPSFsigma[indCh] * dPSFsigma[indCh]);
		nKernelSize[indCh] = (int) Math.ceil(3.0*dPSFsigma[indCh]);
		if(nKernelSize[indCh]%2 == 0)
			 nKernelSize[indCh]++;

		return true;
		
	}
	
	
	/** function adds parameters values to registry **/
	public void setPrefs()
	{
		String sChN = Integer.toString(iCh);
		Prefs.set("ComDet.bBigParticles"+sChN, bBigParticles[iCh-1]);
		Prefs.set("ComDet.bSegmentLargeParticles"+sChN, bSegmentLargeParticles[iCh-1]);

		Prefs.set("ComDet.dPSFsigma"+sChN, dPSFsigma[iCh-1]*2.0);
		Prefs.set("ComDet.dSNRT"+sChN, nSensitivity[iCh-1]);
		
		if(ChNumber == 1)		
		{
			Prefs.set("ComDet.sROIManagerOne", sROIManagerOneCh[nRoiManagerAdd]);
		}
		Prefs.set("ComDet.sSummaryOptions", sSummaryOptions[nSummaryOptions]);
		
		Prefs.set("ComDet.nRoiOption", sRoiOptions[nRoiOption]);		
		
		//Prefs.set("ComDet.bPreview", bPreview);
	}	
	
	
}
